//============================================================================
// Expr2.g - Parse, Create an AST, and Evaluate
//
// Author:  Bary W Pollack
// Date:    Mar. 06, 2004 - 21:00
//
// This example shows how to create an AST and then walk it
// in order to create the corresponding semantics (evaluation)
//
//============================================================================

// Import the necessary classes
{
import java.io.*;
import java.util.*;
import antlr.debug.misc.ASTFrame;
}


//----------------------------------------------------------------------------
// Define a Parser, call it Expr2Parser
//----------------------------------------------------------------------------

class Expr2Parser extends Parser;

options {
    buildAST = true;
}

{
    public static void main(String[] args) throws Exception {
        Expr2Lexer  lexer  = new Expr2Lexer(System.in);
        Expr2Parser parser = new Expr2Parser(lexer);
        parser.expr2();

        // Create an AST for the input expression
        AST tree = parser.getAST();
        System.out.println("Preorder traversal: " + tree.toStringTree());

        // Create an ASTFrame to show the tree structure in 2D
        ASTFrame frame = new ASTFrame("Tree for: " + tree.toStringList(), tree);
        frame.setVisible(true);

        // Walk the AST to create the semantics (evaluate the expression)
        Expr2Walker treeWalker = new Expr2Walker();
        double dValue = treeWalker.expr(tree);

        System.out.println();
        System.out.println("* Success; dValue = " + dValue + " *");
        System.out.println();
    }
}


//==================================================================
// Expr2 Syntax - arithmetic expressions
//==================================================================

/* 
   postfix ! means do NOT create a node for this token
   postfix ^ means make this token the root of the created subtree
*/

expr2    : expr SEMI! ;

expr     : sumExpr ;

sumExpr  : prodExpr ((PLUS^|MINUS^) prodExpr)* ; 

prodExpr : powExpr ((MUL^|DIV^|MOD^) powExpr)* ;

powExpr  : atom (POW^ atom)? ;

atom     : INT | LPAREN! expr RPAREN! ;


//----------------------------------------------------------------------------
// The Expr2 Scanner
//----------------------------------------------------------------------------

class Expr2Lexer extends Lexer;

options {
    charVocabulary = '\0'..'\377';
    k = 2;
}

LPAREN : '(' ;
RPAREN : ')' ;
PLUS   : '+' ;
MINUS  : '-' ;
MUL    : '*' ;
DIV    : '/' ;
MOD    : '%' ;
POW    : '^' ;
SEMI   : ';' ;
// If protected is not specified, there is a nondeterminism error between DIGIT and INT
protected
DIGIT  : '0'..'9' ;
INT    : (DIGIT)+ ;

WS     : ( ' '
         | '\r' '\n'
         | '\n'
         | '\t'
         )
         {$setType(Token.SKIP);}
       ;


//----------------------------------------------------------------------------
// The Expr2 Walker
//----------------------------------------------------------------------------

class Expr2Walker extends TreeParser;

options {
    importVocab = Expr2Parser;
}

/*
  These rule definitions all use the AST syntax:
      #(ROOT child.1 child.2 ... child.n);
    
  For the first production below this says: if PLUS is found as a root, assign
  the values of the evaluation of the child subtrees to the variables 'a' and
  'b' respectively.  The rule does not literally say "evaluate the subtrees;"
  this will happen automatically due to the fact that the rule says to match 
  a tree with PLUS as a root that has two expr's as children.  It is in the 
  matching of these children, in order to match the whole rule, that the 
  subtrees will be evaluated.  The action specified upon a successful match, 
  is to set r equal to a+b. 
*/

expr returns [ double r ]
  { double a, b; r = 0; }

  : #(PLUS  a=expr b=expr)  { r = a+b; }
  | #(MINUS a=expr b=expr)  { r = a-b; }
  | #(MUL   a=expr b=expr)  { r = a*b; }
  | #(DIV   a=expr b=expr)  { r = a/b; }
  | #(MOD   a=expr b=expr)  { r = a%b; }
  | #(POW   a=expr b=expr)  { r = Math.pow(a,b); }
  | i:INT   { r = (double)Integer.parseInt(i.getText()); }
  ;

// EOF
